@@ -2,6 +2,7 @@ package Dancer::Plugin::RPC::XMLRPC;
 use v5.10;
 use Dancer ':syntax';
 use Dancer::Plugin;
+use Scalar::Util 'blessed';
 
 no if $] >= 5.018, warnings => 'experimental::smartmatch';
 
@@ -10,6 +11,7 @@ use Dancer::RPCPlugin::DispatchFromConfig;
 use Dancer::RPCPlugin::DispatchFromPod;
 use Dancer::RPCPlugin::DispatchItem;
 use Dancer::RPCPlugin::DispatchMethodList;
+use Dancer::RPCPlugin::FlattenData;
 
 use Params::Validate ':all';
 use RPC::XML::ParserFactory;
@@ -26,20 +28,12 @@ register xmlrpc => sub {
     given ($arguments->{publish} // 'config') {
         when (exists $dispatch_builder_map{$_}) {
             $publisher = $dispatch_builder_map{$_};
-
             $arguments->{arguments} = plugin_setting() if $_ eq 'config';
         }
         default {
             $publisher = $_;
         }
     }
-
-    my $code_wrapper = $arguments->{code_wrapper} // sub {
-        my $code = shift;
-        my $pkg  = shift;
-        $code->(@_);
-    };
-    my $callback = $arguments->{callback};
     my $dispatcher = $publisher->($arguments->{arguments}, $endpoint);
 
     my $lister = Dancer::RPCPlugin::DispatchMethodList->new();
@@ -49,19 +43,30 @@ register xmlrpc => sub {
         methods  => [ sort keys %{ $dispatcher } ],
     );
 
+    my $code_wrapper = $arguments->{code_wrapper}
+        ? $arguments->{code_wrapper}
+        : sub {
+            my $code = shift;
+            my $pkg  = shift;
+            $code->(@_);
+        };
+    my $callback = $arguments->{callback};
+
+    debug("Starting xmlrpc-handler build: ", $lister);
     my $handle_call = sub {
         if (request->content_type ne 'text/xml') {
             pass();
         }
-        debug("[handle_xmlrpc_call] Processing: ", request->body);
+        debug("[handle_xmlrpc_request] Processing: ", request->body);
 
         local $RPC::XML::ENCODING = $RPC::XML::ENCODING ='UTF-8';
         my $p = RPC::XML::ParserFactory->new();
         my $request = $p->parse(request->body);
-        debug("[handle_xmlrpc_call] Parsed request: ", $request);
-
         my $method_name = $request->name;
+        debug("[handle_xmlrpc_call($method_name)] ", $request->args);
+
         if (! exists $dispatcher->{$method_name}) {
+            warning("$endpoint/#$method_name not found, pass()");
             pass();
         }
 
@@ -81,9 +86,17 @@ register xmlrpc => sub {
             };
             return xmlrpc_response($response);
         }
-        if (! $continue->success) {
-            $response->{faultCode} = $continue->error_code;
-            $response->{faultString} = $continue->error_message;
+        if (!blessed($continue) || !$continue->isa('Dancer::RPCPlugin::CallbackResult')) {
+            $response = {
+                faultCode   => 500,
+                faultString => "Internal error: 'callback_result' wrong class " . blessed($continue),
+            };
+        }
+        elsif (blessed($continue) && !$continue->success) {
+            $response = {
+                faultCode   => $continue->error_code,
+                faultString => $continue->error_message,
+            };
         }
         else {
             my Dancer::RPCPlugin::DispatchItem $di = $dispatcher->{$method_name};
@@ -94,16 +107,24 @@ register xmlrpc => sub {
                 $code_wrapper->($handler, $package, $method_name, @method_args);
             };
 
+            debug("[handling_xmlrpc_response($method_name)] ", $response);
             if (my $error = $@) {
                 $response = {
                     faultCode => 500,
                     faultString => $error,
                 };
             }
+            if (blessed($response) && $response->can('as_xmlrpc_fault')) {
+                $response = $response->as_xmlrpc_fault;
+            }
+            elsif (blessed($response)) {
+                $response = flatten_data($response);
+            }
         }
         return xmlrpc_response($response);
     };
 
+    debug("setting route (xmlrpc): $endpoint ", $lister);
     post $endpoint, $handle_call;
 };
 
